==> bitmat-tl-0.1.0.tar.gz: bitmat-tl-0.1.0/bitmat/utils/bitmat.py <==
from typing import Tuple

import torch

from triton_kernels.bitmat_kernel import batched_bitmat
from utils.packing import pack_ternary


def terniarize(weights: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Terniarizes the weights and returns the scale.
    """
    scale = 1 / torch.max(weights.abs().mean(), torch.tensor(1e-5))
    return torch.clamp((weights * scale).to(torch.int8), -1, 1), scale

def quantize_activations(x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Quantizes the activations and returns the scale for each row.
    """
    scale = (127 / torch.max(x.abs().max(dim=-1).values, torch.tensor(1e-5))).unsqueeze(-1)
    return torch.clamp((x * scale), -127, 128).to(torch.int8), scale


class BitMat(torch.autograd.Function):
    def forward(ctx, W, X):
        """
        During the forward pass, we ternarize the weights, pack them and then quantize the activations.
        We then perform the bit matrix multiplication and return the scaled results.
        ternarization:
        scale_w = 1 / mean(abs(W))                              | STE
        W = clip(round(W * scale_w), -1, 1)                     | STE
        packing:
        packed_w = 4 int8 -> 1 int8                             | STE
        quantization:
        scale_x = 127 / max(abs(X))                             | STE
        X = clip(round(X * scale_x), -127, 128)                 | STE
        bit matrix multiplication:
        Y = X @ w_packed.t()                                    | dot product
        Y = Y / (scale_w * scale_x)                             | STE
        """
        W, scale_w = terniarize(W)
        packed_w = pack_ternary(W, 4)
        X, scale_x = quantize_activations(X)
        ctx.save_for_backward(X, scale_x)
        y = batched_bitmat(X, packed_w)
        return y / scale_w / scale_x

    @staticmethod
    def backward(ctx, grad_output):
        X, scale_x = ctx.saved_tensors
        # get dW
        grad_W = grad_output @ (X / scale_x).t().to(torch.float16) # TODO: check dim
        return grad_W, None

def bitmat(W: torch.Tensor, X: torch.Tensor) -> torch.Tensor:
    return BitMat.apply(W, X)

==> bitmat-tl-0.1.7.tar.gz: bitmat-tl-0.1.7/bitmat/utils/bitmat.py <==
from typing import Tuple

import torch

from ..triton_kernels.bitmat_kernel import batched_bitmat
from .packing import pack_ternary


def terniarize(weights: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Terniarizes the weights and returns the scale.
    """
    scale = 1 / torch.max(weights.abs().mean(), torch.tensor(1e-5))
    return torch.clamp((weights * scale).to(torch.int8), -1, 1), scale

def quantize_activations(x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Quantizes the activations and returns the scale for each row.
    """
    scale = (127 / torch.max(x.abs().max(dim=-1).values, torch.tensor(1e-5))).unsqueeze(-1)
    return torch.clamp((x * scale), -127, 128).to(torch.int8), scale


class BitMat(torch.autograd.Function):
    @staticmethod
    @torch.cuda.amp.custom_fwd
    def forward(ctx, W, X, scale_w=None):
        """
        During the forward pass, we ternarize the weights, pack them and then quantize the activations.
        We then perform the bit matrix multiplication and return the scaled results.
        ternarization:
        scale_w = 1 / mean(abs(W))                              | STE
        W = clip(round(W * scale_w), -1, 1)                     | STE
        packing:
        packed_w = 4 int8 -> 1 int8                             | STE
        quantization:
        scale_x = 127 / max(abs(X))                             | STE
        X = clip(round(X * scale_x), -127, 128)                 | STE
        bit matrix multiplication:
        Y = X @ w_packed.t()                                    | dot product
        Y = Y / (scale_w * scale_x)                             | STE
        """
        if scale_w is None:
            W, scale_w = terniarize(W)
            packed_w = pack_ternary(W, 4)
            ctx.save_for_backward(X)
            X, scale_x = quantize_activations(X)
            y = batched_bitmat(X, packed_w)
            return y / scale_w / scale_x
        else:
            X, scale_x = quantize_activations(X)
            y = batched_bitmat(X, W)
            return y / scale_w / scale_x


    @staticmethod
    @torch.cuda.amp.custom_bwd
    def backward(ctx, grad_output):
        X = ctx.saved_tensors[0]
        # get dW
        grad_W = X.transpose(1, 2) @ grad_output
        return grad_W, None, None

def bitmat(W: torch.Tensor, X: torch.Tensor) -> torch.Tensor:
    return BitMat.apply(W, X)
